import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.TimeZone;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.quartz.Job;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;

/**
 * Basic data collection job, implementing basic configuration handling and collection routines.
 * 
 * @author fhogenboom
 */
public abstract class CollectJob implements Job
{

	protected ArrayList<Object> hday;
	protected Integer           hour_stop, hour_strt, mnte_stop, mnte_strt, scnd_stop, scnd_strt;
	protected String            date, form_date, form_hday, form_time, time, time_zone;

	private static Logger       _log = LoggerFactory.getLogger(CollectJob.class);

	/**
	 * Constructor.
	 */
	public CollectJob()
	{
		super();
	}

	/**
	 * Configures the {@link org.quartz.JobDataMap CollectJob} component by parsing a configuration
	 * file. Configurations are stored in a {@link org.quartz.JobDataMap}, which is available during
	 * runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link Boolean} value indicating whether parsed configuration is valid.
	 */
	private boolean configure(JobExecutionContext context)
	{
		// Load data map of current component
		JobDataMap dataMap = Util.loadDataMap(context);

		// If component has not yet been configured
		if (!dataMap.getBooleanValue("configured"))
		{
			// Configure
			getLogger().info("Configuring");
			boolean c_shared = configureShared();
			boolean c_local = configureLocal(context.getJobDetail().getKey().getName());
			dataMap.put("configured", true);
			Util.storeDataMap(context, dataMap);
			store(context);
			// Return whether parsed configuration is valid
			return (c_shared && c_local);
		}
		// Else
		else
		{
			// Return that parsed configuration is valid
			return true;
		}
	}

	/**
	 * Configures the {@link CollectJob} component by parsing a configuration file for shared market
	 * elements.
	 * 
	 * @return {@link Boolean} value indicating whether parsed configuration is valid.
	 */
	@SuppressWarnings("unchecked")
	private boolean configureShared()
	{
		try
		{
			// Load market configuration
			hour_strt = Util.getInteger(Util.parseXMLelement("market", "hour_strt"));
			hour_stop = Util.getInteger(Util.parseXMLelement("market", "hour_stop"));
			mnte_strt = Util.getInteger(Util.parseXMLelement("market", "mnte_strt"));
			mnte_stop = Util.getInteger(Util.parseXMLelement("market", "mnte_stop"));
			scnd_strt = Util.getInteger(Util.parseXMLelement("market", "scnd_strt"));
			scnd_stop = Util.getInteger(Util.parseXMLelement("market", "scnd_stop"));
			form_date = Util.getString(Util.parseXMLelement("market", "form_date"));
			form_time = Util.getString(Util.parseXMLelement("market", "form_time"));
			form_hday = Util.getString(Util.parseXMLelement("market", "form_hday"));
			time_zone = Util.getString(Util.parseXMLelement("market", "time_zone"));
			hday = (ArrayList<Object>) Util.parseXMLelements("market", "holidays");
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Return whether parsed configuration is valid
		return (hour_strt != null && hour_stop != null && mnte_strt != null && mnte_stop != null
		        && scnd_strt != null && scnd_stop != null && form_date != null && form_time != null
		        && form_hday != null && time_zone != null && hday != null);
	}

	/**
	 * Configures the {@link CollectJob} component by loading configurations stored in a
	 * {@link org.quartz.JobDataMap}, which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link Boolean} value indicating whether loaded configuration is valid.
	 */
	private boolean load(JobExecutionContext context)
	{
		// Load configuration
		boolean l_shared = loadShared(context);
		boolean l_local = loadLocal(context);
		// Return whether loaded configuration is valid
		return (l_shared && l_local);
	}

	/**
	 * Configures the {@link CollectJob} component by loading configurations for shared market
	 * elements stored in a {@link org.quartz.JobDataMap}, which is available during runtime of the
	 * component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link Boolean} value indicating whether loaded configuration is valid.
	 */
	@SuppressWarnings("unchecked")
	private boolean loadShared(JobExecutionContext context)
	{
		try
		{
			// Load data map of current component
			JobDataMap dataMap = Util.loadDataMap(context);

			// Load market configuration
			hour_strt = new Integer(dataMap.getInt("hour_strt"));
			hour_stop = new Integer(dataMap.getInt("hour_stop"));
			mnte_strt = new Integer(dataMap.getInt("mnte_strt"));
			mnte_stop = new Integer(dataMap.getInt("mnte_stop"));
			scnd_strt = new Integer(dataMap.getInt("scnd_strt"));
			scnd_stop = new Integer(dataMap.getInt("scnd_stop"));
			form_date = dataMap.getString("form_date");
			form_time = dataMap.getString("form_time");
			form_hday = dataMap.getString("form_hday");
			time_zone = dataMap.getString("time_zone");
			hday = (ArrayList<Object>) dataMap.get("hday");
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Return whether loaded configuration is valid
		return (hour_strt != null && hour_stop != null && mnte_strt != null && mnte_stop != null
		        && scnd_strt != null && scnd_stop != null && form_date != null && form_time != null
		        && form_hday != null && time_zone != null && hday != null);
	}

	/**
	 * Stores the configuration of the {@link CollectJob} component in a
	 * {@link org.quartz.JobDataMap}, which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 */
	private void store(JobExecutionContext context)
	{
		// Store configuration
		storeShared(context);
		storeLocal(context);
	}

	/**
	 * Stores the configuration for shared market elements of the {@link CollectJob} component in a
	 * {@link org.quartz.JobDataMap}, which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 */
	private void storeShared(JobExecutionContext context)
	{
		try
		{
			// Load data map of current component
			JobDataMap dataMap = Util.loadDataMap(context);

			// Store market configuration
			dataMap.put("hour_strt", hour_strt);
			dataMap.put("hour_stop", hour_stop);
			dataMap.put("mnte_strt", mnte_strt);
			dataMap.put("mnte_stop", mnte_stop);
			dataMap.put("scnd_strt", scnd_strt);
			dataMap.put("scnd_stop", scnd_stop);
			dataMap.put("form_date", form_date);
			dataMap.put("form_time", form_time);
			dataMap.put("form_hday", form_hday);
			dataMap.put("time_zone", time_zone);
			dataMap.put("hday", hday);

			// Store data map of current component
			Util.storeDataMap(context, dataMap);
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
	}

	/**
	 * Executes the tasks of the {@link CollectJob} component.
	 * 
	 * @param context Execution context of the component.
	 * @throws JobExecutionException An exception that indicates to the scheduler that an error
	 *             occurred while executing, and whether or not the component requests to be
	 *             re-fired immediately, or whether it wants to be unscheduled.
	 */
	public void execute(JobExecutionContext context) throws JobExecutionException
	{
		try
		{
			// If parsed configuration is invalid
			if (!configure(context))
			{
				throw new Exception("Configuration error");
			}
			// Else
			else
			{
				// If loaded configuration is valid
				if (load(context))
				{
					// Create time stamp
					Calendar date_now = Calendar.getInstance();
					date_now.setTimeZone(TimeZone.getTimeZone(time_zone));
					DateFormat sdf1 = new SimpleDateFormat(form_date);
					sdf1.setTimeZone(TimeZone.getTimeZone(time_zone));
					DateFormat sdf2 = new SimpleDateFormat(form_time);
					sdf2.setTimeZone(TimeZone.getTimeZone(time_zone));
					date = sdf1.format(date_now.getTime());
					time = sdf2.format(date_now.getTime());
					// If market is open
					if (isMarket(date_now))
					{
						// Execute data collection
						String jobKey = context.getJobDetail().getKey().getName();
						getLogger().info(
						        "Executing " + jobKey + " at local time " + date + " " + time);
						collect();
					}
				}
				// Store configuration
				store(context);
			}
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
	}

	/**
	 * Checks whether the market is open.
	 * 
	 * @param date_now The date to be checked.
	 * @return {@link Boolean} value indicating whether the market is open.
	 */
	private boolean isMarket(Calendar date_now)
	{
		if (!isWeekend(date_now) & !isHoliday(date_now) & isTime(date_now))
		{
			return true;
		}
		return false;
	}

	/**
	 * Checks whether it is a weekend.
	 * 
	 * @param date_now The date to be checked.
	 * @return {@link Boolean} value indicating whether it is a weekend.
	 */
	private boolean isWeekend(Calendar date_now)
	{
		return (date_now.get(Calendar.DAY_OF_WEEK) == Calendar.SATURDAY | date_now
		        .get(Calendar.DAY_OF_WEEK) == Calendar.SUNDAY);
	}

	/**
	 * Checks whether it is a holiday.
	 * 
	 * @param date_now The date to be checked.
	 * @return {@link Boolean} value indicating whether it is a holiday.
	 */
	private boolean isHoliday(Calendar date_now)
	{
		DateFormat sdf = new SimpleDateFormat(form_hday);
		sdf.setTimeZone(TimeZone.getTimeZone(time_zone));
		boolean isHoliday = false;
		Iterator<Object> it = hday.iterator();
		while (it.hasNext() && !isHoliday)
		{
			try
			{
				Calendar date_hol = Calendar.getInstance();
				date_hol.setTimeZone(TimeZone.getTimeZone(time_zone));
				date_hol.setTime(sdf.parse((String) it.next()));
				if (date_now.get(Calendar.YEAR) == date_hol.get(Calendar.YEAR)
				        & date_now.get(Calendar.MONTH) == date_hol.get(Calendar.MONTH)
				        & date_now.get(Calendar.DAY_OF_MONTH) == date_hol
				                .get(Calendar.DAY_OF_MONTH))
				{
					isHoliday = true;
				}
			}
			catch (Exception e)
			{
				getLogger().error(e.toString());
			}
		}
		return isHoliday;
	}

	/**
	 * Checks whether it is a trading time.
	 * 
	 * @param date_now The date to be checked.
	 * @return {@link Boolean} value indicating whether it is a trading time.
	 */
	private boolean isTime(Calendar date_now)
	{
		return (date_now.get(Calendar.HOUR_OF_DAY) == hour_strt
		        & date_now.get(Calendar.MINUTE) == mnte_strt & date_now.get(Calendar.SECOND) >= scnd_strt)
		        | (date_now.get(Calendar.HOUR_OF_DAY) == hour_strt & date_now.get(Calendar.MINUTE) > mnte_strt)
		        | (date_now.get(Calendar.HOUR_OF_DAY) > hour_strt & date_now
		                .get(Calendar.HOUR_OF_DAY) < hour_stop)
		        | (date_now.get(Calendar.HOUR_OF_DAY) == hour_stop
		                & date_now.get(Calendar.MINUTE) == mnte_stop & date_now
		                .get(Calendar.SECOND) == scnd_stop);
	}

	/**
	 * Collects data from a specific source.
	 */
	protected abstract void collect();

	/**
	 * Configures the {@link CollectJob} component by parsing a configuration file for local
	 * component elements.
	 * 
	 * @param jobKey ID of the component.
	 * @return {@link Boolean} value indicating whether parsed configuration is valid.
	 */
	protected abstract boolean configureLocal(String jobKey);

	/**
	 * Configures the {@link CollectJob} component by loading configurations for local component
	 * elements stored in a {@link org.quartz.JobDataMap}, which is available during runtime of the
	 * component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link Boolean} value indicating whether loaded configuration is valid.
	 */
	protected abstract boolean loadLocal(JobExecutionContext context);

	/**
	 * Retrieves the {@link org.slf4j.Logger} of the component.
	 * 
	 * @return {@link org.slf4j.Logger} used for logging system output.
	 */
	protected abstract Logger getLogger();

	/**
	 * Stores the configuration for local component elements of the {@link CollectJob} component in
	 * a {@link org.quartz.JobDataMap}, which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 */
	protected abstract void storeLocal(JobExecutionContext context);

}
