import java.util.ArrayList;
import java.util.Iterator;

import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Basic stock collection job, implementing basic configuration handling and collection routines.
 * 
 * @author fhogenboom
 */
public abstract class StockJob extends CollectJob
{

	protected ArrayList<Object> comp, stock_xml_elms;
	protected Database          dbs;
	protected Integer           max_jobs, num_files, num_jobs, zip_buffer, zip_files;
	protected String            date_now, file_path, file_ext, job_key, stock_name, stock_txt_date,
	        stock_txt_time, stock_url, stock_url_base, stock_url_parm, stock_xml_attr,
	        stock_xml_root;

	private static boolean      zipping   = false;
	private static int          num_parts = 0, num_out = 0;
	private static String       cur_date  = null;

	private static Logger       _log      = LoggerFactory.getLogger(StockJob.class);

	/**
	 * Constructor.
	 */
	public StockJob()
	{
		super();
	}

	/**
	 * Configures the {@link StockJob} component by parsing a configuration file for local component
	 * elements.
	 * 
	 * @param job_key ID of the component.
	 * @return {@link Boolean} value indicating whether parsed configuration is valid.
	 */
	@SuppressWarnings("unchecked")
	protected boolean configureLocal(String job_key)
	{
		try
		{
			// Load stock configuration
			max_jobs = Util.getInteger(Util.parseXMLelement("stock", "max_jobs"));
			file_path = Util.getString(Util.parseXMLelement("stock", "file_path"));
			file_ext = Util.getString(Util.parseXMLelement("stock", "file_ext"));
			stock_txt_date = Util.getString(Util.parseXMLelement("stock", "stock_txt_date"));
			stock_txt_time = Util.getString(Util.parseXMLelement("stock", "stock_txt_time"));
			zip_files = Util.getInteger(Util.parseXMLelement("stock", "zip_files"));
			zip_buffer = Util.getInteger(Util.parseXMLelement("stock", "zip_buffer"));
			comp = (ArrayList<Object>) Util.parseXMLelements("stock", "companies");

			// Load component configuration
			stock_name = Util.getString(Util.parseXMLid(job_key, "eye", "stock_name"));
			stock_url_base = Util.getString(Util.parseXMLid(job_key, "eye", "stock_url_base"));
			stock_url_parm = Util.getString(Util.parseXMLid(job_key, "eye", "stock_url_parm"));
			stock_xml_attr = Util.getString(Util.parseXMLid(job_key, "eye", "stock_xml_attr"));
			stock_xml_root = Util.getString(Util.parseXMLid(job_key, "eye", "stock_xml_root"));
			stock_xml_elms = (ArrayList<Object>) Util.parseXMLids(job_key, "eye", "stock_xml_elms");
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Return whether parsed configuration is valid
		return (max_jobs != null && file_path != null && file_ext != null && stock_txt_date != null
		        && stock_txt_time != null && zip_files != null && zip_buffer != null
		        && comp != null && stock_name != null && stock_url_base != null
		        && stock_url_parm != null && stock_xml_attr != null && stock_xml_root != null && stock_xml_elms != null);
	}

	/**
	 * Configures the {@link StockJob} component by loading configurations for local component
	 * elements stored in a {@link org.quartz.JobDataMap}, which is available during runtime of the
	 * component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link Boolean} value indicating whether loaded configuration is valid.
	 */
	@SuppressWarnings("unchecked")
	protected boolean loadLocal(JobExecutionContext context)
	{
		try
		{
			// Load data map of current component
			JobDataMap dataMap = Util.loadDataMap(context);

			// Load job key of current component
			job_key = context.getJobDetail().getKey().getName();

			// Load stock configuration
			max_jobs = dataMap.getInt("max_jobs");
			file_path = dataMap.getString("file_path");
			file_ext = dataMap.getString("file_ext");
			stock_txt_date = dataMap.getString("stock_txt_date");
			stock_txt_time = dataMap.getString("stock_txt_time");
			zip_files = dataMap.getInt("zip_files");
			zip_buffer = dataMap.getInt("zip_buffer");
			comp = (ArrayList<Object>) dataMap.get("comp");

			// Load component configuration
			stock_name = dataMap.getString("stock_name");
			stock_url_base = dataMap.getString("stock_url_base");
			stock_url_parm = dataMap.getString("stock_url_parm");
			stock_xml_attr = dataMap.getString("stock_xml_attr");
			stock_xml_root = dataMap.getString("stock_xml_root");
			stock_xml_elms = (ArrayList<Object>) dataMap.get("stock_xml_elms");

			// Load data
			dbs = (Database) dataMap.get("dbs");
			try
			{
				num_files = dataMap.getInt("num_files");
			}
			catch (Exception e)
			{
				num_files = 0;
			}
			try
			{
				num_jobs = dataMap.getInt("num_jobs");
			}
			catch (Exception e)
			{
				num_jobs = 0;
			}
			date_now = dataMap.getString("date_now");
			stock_url = dataMap.getString("stock_url");
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
		// Return whether loaded configuration is valid
		return (max_jobs != null && file_path != null && file_ext != null && stock_txt_date != null
		        && stock_txt_time != null && zip_files != null && zip_buffer != null
		        && comp != null && stock_name != null && stock_url_base != null
		        && stock_url_parm != null && stock_xml_attr != null && stock_xml_root != null && stock_xml_elms != null);
	}

	/**
	 * Stores the configuration for local component elements of the {@link StockJob} component in a
	 * {@link org.quartz.JobDataMap}, which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 */
	protected void storeLocal(JobExecutionContext context)
	{
		try
		{
			// Load data map of current component
			JobDataMap dataMap = Util.loadDataMap(context);

			// Store stock configuration
			dataMap.put("max_jobs", max_jobs);
			dataMap.put("file_path", file_path);
			dataMap.put("file_ext", file_ext);
			dataMap.put("stock_txt_date", stock_txt_date);
			dataMap.put("stock_txt_time", stock_txt_time);
			dataMap.put("zip_files", zip_files);
			dataMap.put("zip_buffer", zip_buffer);
			dataMap.put("comp", comp);

			// Store component configuration
			dataMap.put("stock_name", stock_name);
			dataMap.put("stock_url_base", stock_url_base);
			dataMap.put("stock_url_parm", stock_url_parm);
			dataMap.put("stock_xml_attr", stock_xml_attr);
			dataMap.put("stock_xml_root", stock_xml_root);
			dataMap.put("stock_xml_elms", stock_xml_elms);

			// Store data
			dataMap.put("dbs", dbs);
			dataMap.put("num_files", num_files);
			dataMap.put("num_jobs", num_jobs);
			dataMap.put("date_now", date_now);
			dataMap.put("stock_url", stock_url);

			// Store data map of current component
			Util.storeDataMap(context, dataMap);
		}
		catch (Exception e)
		{
			getLogger().error(e.toString());
		}
	}

	/**
	 * Collects data from a specific source.
	 */
	protected void collect()
	{
		// If URL has not yet been constructed
		if (stock_url == null)
		{
			// Construct URL
			stock_url = stock_url_base + "?" + stock_url_parm + "=";
			Iterator<Object> itc = comp.iterator();
			while (itc.hasNext())
			{
				stock_url += (String) itc.next()
				        + ((itc.hasNext()) ? "&" + stock_url_parm + "=" : "");
			}
			getLogger().info("Fetching data from " + stock_url);

			// Create database
			dbs = new Database(job_key);
			dbs.init();
			dbs.addField(stock_txt_date);
			dbs.addField(stock_txt_time);
			Iterator<Object> ite = stock_xml_elms.iterator();
			while (ite.hasNext())
			{
				dbs.addField((String) ite.next());
			}
		}

		// Save database to file
		saveToFile();

		// Zip database
		zipFiles();

		// Parse feed
		parseFeed();

		// Update counter
		num_jobs++;
	}

	/**
	 * Saves an individual {@link StockJob} component's database to a CSV formatted file whenever
	 * there is a date swap or the maximum number of jobs allowed to be stored in memory (specified
	 * in the configuration file) is exceeded. File path and extension of the file to be saved are
	 * specified in the configuration file. File name is constructed as {@code date_now}
	 * {@code stock_name} {@code num_files}, where {@code date_now} represents the current date
	 * (which can be yesterday's date in case of a date swap), {@code stock_name} represents the
	 * name of a feed, and where {@code num_files} represents the number number of files that have
	 * been created by the individual {@link StockJob} component.
	 */
	private void saveToFile()
	{
		// If current date has not yet been initialized
		if (date_now == null)
		{
			// Initialize
			date_now = date;
		}
		// Else
		else
		{
			// If date swap detected or if memory is full
			if (!date_now.equals(date) || num_jobs >= max_jobs)
			{
				// If database is not empty
				if (num_jobs > 0)
				{
					// Compose file name
					String file = date_now + " " + stock_name + " " + ++num_files;
					// Save file
					dbs.saveToFile(file_path, file, file_ext);
					// Empty database
					dbs.flush();
					// Update output counter
					num_out++;
				}
				// If date swap detected
				if (!date_now.equals(date))
				{
					// Reset file counter
					num_files = 0;
				}
				// Reset job counter
				num_jobs = 0;
				// Update date
				date_now = date;
			}
		}
	}

	/**
	 * Zips output files whenever there is a date swap or the maximum number of files allowed in a
	 * single zip file (specified in the configuration file) is exceeded. File path and extension of
	 * the files to be zipped are specified in the configuration file, as well as the buffer used
	 * for zipping. Zip files are stored in the same directory as the files they contain. After
	 * completion, target files are removed. File name is constructed as {@code cur_date}-
	 * {@code num_parts}, where {@code cur_date} represents the current date (which can be
	 * yesterday's date in case of a date swap) and where {@code num_parts} represents the number
	 * number of zip files that have been created by all {@link StockJob} components.
	 */
	private void zipFiles()
	{
		// If current date is not specified
		if (cur_date == null)
		{
			// Update date
			cur_date = date;
		}
		// Else
		else
		{
			// If date swap detected or if the maximum number of files allowed in a single zip
			// file is exceeded
			if (!cur_date.equals(date) || num_out >= zip_files)
			{
				// If no NewsJob component is already zipping files
				if (!zipping)
				{
					// Add zip lock
					zipping = true;
					// If there are files to zip
					if (num_out > 0)
					{
						// Zip files
						Util.zipFiles(getClass().getSuperclass().getName(), file_path, cur_date
						        + "-" + ++num_parts, file_ext, zip_buffer, zip_files);
					}
					// If there is a date swap
					if (!cur_date.equals(date))
					{
						// Reset counter
						num_parts = 0;
					}
					// Update files left to be zipped
					num_out -= Math.min(num_out, zip_files);
					// Update date
					cur_date = date;
					// Remove zip lock
					zipping = false;
				}
			}
		}
	}

	/**
	 * Retrieves the {@link org.slf4j.Logger} of the component.
	 * 
	 * @return {@link org.slf4j.Logger} used for logging system output.
	 */
	protected abstract Logger getLogger();

	/**
	 * Parses feed containing stock data.
	 */
	protected abstract void parseFeed();

}
