import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.WildcardFileFilter;

import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Contains tools useful for XML-based configuration file handling as well as file zipping.
 * 
 * @author fhogenboom
 * 
 */
public class Util
{

	private static HashMap<String, JobDataMap> config     = new HashMap<String, JobDataMap>();
	private static int                         writers    = 0;
	private static String                      config_src = null;

	private static Logger                      _log       = LoggerFactory.getLogger(Util.class);

	/**
	 * Constructor.
	 */
	public Util()
	{
		super();
	}

	/**
	 * Sets location of the configuration file to be used.
	 * 
	 * @param src Path and file name of configuration file to be used.
	 */
	public static void setConfig(String src)
	{
		config_src = src;
	}

	/**
	 * Loads configurations for local component elements stored in a {@link org.quartz.JobDataMap},
	 * which is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 * @return {@link JobDataMap} representing component configuration.
	 */
	public static JobDataMap loadDataMap(JobExecutionContext context)
	{
		JobDataMap dataMap = config.get(context.getJobDetail().getKey().getName());
		if (dataMap == null)
		{
			dataMap = context.getJobDetail().getJobDataMap();
		}
		return dataMap;
	}

	/**
	 * Stores configurations for local component elements in a {@link org.quartz.JobDataMap}, which
	 * is available during runtime of the component.
	 * 
	 * @param context Execution context of the component.
	 * @param dataMap Component configuration.
	 */
	public static void storeDataMap(JobExecutionContext context, JobDataMap dataMap)
	{
		config.put(context.getJobDetail().getKey().getName(), dataMap);
	}

	/**
	 * Parses an {@link Object} to an {@link Integer}.
	 * 
	 * @param object {@link Object} to be parsed.
	 * @return {@link Integer} value representing the {@link Object} to be parsed.
	 */
	public static Integer getInteger(Object object)
	{
		if (object != null)
		{
			return Integer.parseInt((String) object);
		}
		else
		{
			return null;
		}
	}

	/**
	 * Parses an {@link Object} to a {@link String}.
	 * 
	 * @param object {@link Object} to be parsed.
	 * @return {@link String} value representing the {@link Object} to be parsed.
	 */
	public static String getString(Object object)
	{
		if (object != null)
		{
			return (String) object;
		}
		else
		{
			return null;
		}
	}

	/**
	 * Parses an XML-based configuration file and looks for one or more values of a (sub)element.
	 * 
	 * @param element1 Main element to be parsed.
	 * @param element2 Sub-element to be parsed.
	 * @param list {@link Boolean} value indicating whether one ({@code false}) or multiple (
	 *            {@code true}) values are desired
	 * @return {@link Object} representing the value(s) of the (sub)element to be parsed. Depending
	 *         on the value of {@code list} this is either an {@link Object} or an {@link ArrayList}
	 *         of {@link Object} instances.
	 */
	private static Object parseXMLelement(String element1, String element2, boolean list)
	{
		_log.debug("Parsing configuration file " + config_src + " for " + element1 + " > "
		        + element2);
		ArrayList<Object> values = new ArrayList<Object>();
		Object value = null;
		try
		{
			// Open file
			File file = new File(config_src);
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			// Parse XML document
			Document doc = db.parse(file);
			doc.getDocumentElement().normalize();
			// Get root element
			NodeList roots = doc.getElementsByTagName(element1);
			if (roots.getLength() == 0)
			{
				throw new Exception("Unable to locate element " + element1);
			}
			// Get children
			NodeList root = roots.item(0).getChildNodes();
			// If a list is desired
			if (list)
			{
				NodeList children = null;
				// For each child
				for (int i = 0; i < root.getLength() && children == null; i++)
				{
					Node element = root.item(i);
					// If child is equal to the sub-element to be found
					if (element.getNodeName().equals(element2))
					{
						// Store children of child element
						children = element.getChildNodes();
					}
				}
				if (children == null)
				{
					throw new Exception("Unable to locate element " + element1 + " > " + element2);
				}
				// For each child
				for (int i = 0; i < children.getLength(); i++)
				{
					Node element = children.item(i);
					// If child is equal to item
					if (element.getNodeName().equals("item"))
					{
						// Store textual content in ArrayList
						values.add(element.getTextContent());
					}
				}
			}
			// Else
			else
			{
				// For each child
				for (int i = 0; i < root.getLength() && value == null; i++)
				{
					Node element = root.item(i);
					// If child is equal to the sub-element to be found
					if (element.getNodeName().equals(element2))
					{
						// Store textual content in Object
						value = element.getTextContent();
					}
				}
				if (value == null)
				{
					throw new Exception("Unable to locate element " + element1 + " > " + element2);
				}
			}
		}
		catch (Exception e)
		{
			_log.error(e.toString());
			values = null;
		}
		return (list) ? values : value;
	}

	/**
	 * Parses an XML-based configuration file and looks for the value of a (sub)element.
	 * 
	 * @param element1 Main element to be parsed.
	 * @param element2 Sub-element to be parsed.
	 * @return {@link Object} representing the value of the (sub)element to be parsed.
	 */
	public static Object parseXMLelement(String element1, String element2)
	{
		return parseXMLelement(element1, element2, false);
	}

	/**
	 * Parses an XML-based configuration file and looks for multiple values of a (sub)element.
	 * 
	 * @param element1 Main element to be parsed.
	 * @param element2 Sub-element to be parsed.
	 * @return {@link Object} representing the values of the (sub)element to be parsed.
	 */
	public static Object parseXMLelements(String element1, String element2)
	{
		return parseXMLelement(element1, element2, true);
	}

	/**
	 * Parses an XML-based configuration file and looks for one or more values of a (sub)element
	 * with a specific ID.
	 * 
	 * @param id Unique id of the (sub)element to be parsed.
	 * @param element1 Main element to be parsed.
	 * @param element2 Sub-element to be parsed.
	 * @param list {@link Boolean} value indicating whether one ({@code false}) or multiple (
	 *            {@code true}) values are desired
	 * @return {@link Object} representing the value(s) of the (sub)element to be parsed. Depending
	 *         on the value of {@code list} this is either an {@link Object} or an {@link ArrayList}
	 *         of {@link Object} instances.
	 */
	private static Object parseXMLid(String id, String element1, String element2, boolean list)
	{
		_log.debug("Parsing configuration file " + config_src + " for " + element1 + " (id: " + id
		        + ") > " + element2);
		ArrayList<Object> values = new ArrayList<Object>();
		Object value = null;
		try
		{
			// Open file
			File file = new File(config_src);
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			// Parse XML document
			Document doc = db.parse(file);
			doc.getDocumentElement().normalize();
			// Get root elements
			NodeList roots = doc.getElementsByTagName(element1);
			if (roots.getLength() == 0)
			{
				throw new Exception("Unable to locate element " + element1);
			}
			NodeList root = null;
			// For each element
			for (int i = 0; i < roots.getLength() && root == null; i++)
			{
				Node element = roots.item(i);
				// Get element id
				Node par = element.getAttributes().getNamedItem("id");
				if (par == null)
				{
					throw new Exception("Unable to locate parameter \"id\" element " + element1);
				}
				// If element id equals the id to be found
				if (par.getNodeValue().equals(id))
				{
					// Store element
					root = element.getChildNodes();
				}
			}
			if (root == null || root.getLength() == 0)
			{
				throw new Exception("Unable to locate element " + element1 + " (id: " + id + ")");
			}
			// If a list is desired
			if (list)
			{
				NodeList children = null;
				// For each child
				for (int i = 0; i < root.getLength() && children == null; i++)
				{
					Node element = root.item(i);
					// If child is equal to the sub-element to be found
					if (element.getNodeName().equals(element2))
					{
						// Store children of child element
						children = element.getChildNodes();
					}
				}
				if (children == null)
				{
					throw new Exception("Unable to locate element " + element1 + " (id: " + id
					        + ") > " + element2);
				}
				// For each child
				for (int i = 0; i < children.getLength(); i++)
				{
					Node element = children.item(i);
					// If child is equal to item
					if (element.getNodeName().equals("item"))
					{
						// Store textual content in ArrayList
						values.add(element.getTextContent());
					}
				}
			}
			// Else
			else
			{
				// For each child
				for (int i = 0; i < root.getLength() && value == null; i++)
				{
					Node element = root.item(i);
					// If child is equal to the sub-element to be found
					if (element.getNodeName().equals(element2))
					{
						// Store textual content in Object
						value = element.getTextContent();
					}
				}
				if (value == null)
				{
					throw new Exception("Unable to locate element " + element1 + " (id: " + id
					        + ") > " + element2);
				}
			}
		}
		catch (Exception e)
		{
			_log.error(e.toString());
			values = null;
		}
		return (list) ? values : value;
	}

	/**
	 * Parses an XML-based configuration file and looks for the value of a (sub)element with a
	 * specific ID.
	 * 
	 * @param id Unique id of the (sub)element to be parsed.
	 * @param element1 Main element to be parsed.
	 * @param element2 Sub-element to be parsed.
	 * @return {@link Object} representing the value of the (sub)element to be parsed.
	 */
	public static Object parseXMLid(String id, String element1, String element2)
	{
		return parseXMLid(id, element1, element2, false);
	}

	/**
	 * Parses an XML-based configuration file and looks for multiple values of a (sub)element with a
	 * specific ID.
	 * 
	 * @param id Unique id of the (sub)element to be parsed.
	 * @param element1 Main element to be parsed.
	 * @param element2 Sub-element to be parsed.
	 * @return {@link Object} representing the values of the (sub)element to be parsed.
	 */
	public static Object parseXMLids(String id, String element1, String element2)
	{
		return parseXMLid(id, element1, element2, true);
	}

	/**
	 * Parses an XML-based configuration file and looks for the values of the parameters of an
	 * element.
	 * 
	 * @param element1 Main element to be parsed.
	 * @param parameter1 First parameter to be parsed.
	 * @param parameter2 Second parameter to be parsed.
	 * @return {@link TreeMap} of {@link Object} instances representing the values of the parameters
	 *         of the element to be parsed.
	 */
	public static TreeMap<Object, Object> parseXMLparameter(String element1, String parameter1,
	        String parameter2)
	{
		_log.debug("Parsing configuration file " + config_src + " for " + element1 + " ("
		        + parameter1 + ", " + parameter2 + ")");
		TreeMap<Object, Object> values = new TreeMap<Object, Object>();
		try
		{
			// Open file
			File file = new File(config_src);
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			DocumentBuilder db = dbf.newDocumentBuilder();
			// Parse XML document
			Document doc = db.parse(file);
			doc.getDocumentElement().normalize();
			// Get root elements
			NodeList roots = doc.getElementsByTagName(element1);
			if (roots.getLength() == 0)
			{
				throw new Exception("Unable to locate element " + element1);
			}
			// For each element
			for (int i = 0; i < roots.getLength(); i++)
			{
				Node element = roots.item(i);
				// Get first parameter
				Node par1 = element.getAttributes().getNamedItem(parameter1);
				if (par1 == null)
				{
					throw new Exception("Unable to locate parameter \"" + parameter1 + "\" element"
					        + element1);
				}
				// Get second parameter
				Node par2 = element.getAttributes().getNamedItem(parameter2);
				if (par2 == null)
				{
					throw new Exception("Unable to locate parameter \"" + parameter2 + "\" element"
					        + element1);
				}
				// Store parameter values in TreeMap
				values.put(par1.getNodeValue(), par2.getNodeValue());
			}
		}
		catch (Exception e)
		{
			_log.error(e.toString());
			values = null;
		}
		return values;
	}

	/**
	 * Zips a file of a given name and extension in a given folder. After completion, target file is
	 * removed. File name is equal to the name of the file to be zipped.
	 * 
	 * @param name Name of the storage initiating zip action.
	 * @param path Path of the file to be zipped.
	 * @param file Name of the file to be zipped.
	 * @param ext Extension of the file to be zipped.
	 * @param buffer Buffer used when zipping.
	 */
	public static void zipFile(String name, String path, String file, String ext, int buffer)
	{
		try
		{
			// Add zip lock
			zipLock();
			_log.info("Zipping storage " + name + " to file " + path + file + ".zip");
			try
			{
				// Open zip stream
				ZipOutputStream out = new ZipOutputStream(new BufferedOutputStream(
				        new FileOutputStream(path + file + ".zip")));
				byte[] data = new byte[buffer];
				// Open input stream
				BufferedInputStream in = new BufferedInputStream(new FileInputStream(path + file
				        + ext));
				out.putNextEntry(new ZipEntry(file + ext));
				int count;
				// Write target file to zip
				while ((count = in.read(data, 0, buffer)) != -1)
				{
					out.write(data, 0, count);
				}
				// Close streams
				in.close();
				out.flush();
				out.close();
				// Delete target file
				try
				{
					new File(path + file + ext).delete();
				}
				catch (Exception e)
				{
					_log.error(e.toString());
				}
			}
			catch (Exception e)
			{
				_log.error(e.toString());
			}
		}
		catch (Exception e)
		{
			_log.error(e.toString());
		}
		finally
		{
			// Remove zip lock
			zipUnlock();
		}
	}

	/**
	 * Zips files of a given extension in a given folder to a single file with a given name. After
	 * completion, target files are removed.
	 * 
	 * @param name Name of the storage initiating zip action.
	 * @param path Path of the files to be zipped.
	 * @param file Name of the resulting zip file.
	 * @param ext Extension of the files to be zipped.
	 * @param buffer Buffer used when zipping.
	 * @param max_files Maximum number of files to be added to the zip file.
	 */
	public static void zipFiles(String name, String path, String file, String ext, int buffer,
	        int max_files)
	{
		try
		{
			// Add zip lock
			zipLock();
			_log.info("Zipping storage " + name + " to file " + path + file + ".zip");
			try
			{
				// Open zip stream
				ZipOutputStream out = new ZipOutputStream(new BufferedOutputStream(
				        new FileOutputStream(path + file + ".zip")));
				byte[] data = new byte[buffer];
				// Create traget file index
				Collection<File> files = FileUtils.listFiles(new File(path),
				        new WildcardFileFilter("*" + ext), null);
				Iterator<File> itFiles = files.iterator();
				int num_files = 0;
				// For each file (up to the allowed maximum number of files)
				while (itFiles.hasNext() && num_files < max_files)
				{
					File comp = itFiles.next();
					// Open input stream
					BufferedInputStream in = new BufferedInputStream(new FileInputStream(
					        comp.getCanonicalPath()));
					out.putNextEntry(new ZipEntry(comp.getName()));
					int count;
					// Write target file to zip
					while ((count = in.read(data, 0, buffer)) != -1)
					{
						out.write(data, 0, count);
					}
					// Close streams
					out.closeEntry();
					in.close();
					// Delete target file
					try
					{
						comp.delete();
					}
					catch (Exception e)
					{
						_log.error(e.toString());
					}
					// Increase counter
					num_files++;
				}
				// Close stream
				out.flush();
				out.close();
			}
			catch (Exception e)
			{
				_log.error(e.toString());
			}
		}
		catch (Exception e)
		{
			_log.error(e.toString());
		}
		finally
		{
			// Remove zip lock
			zipUnlock();
		}
	}

	/**
	 * Locks the zipping process in order to prevent asynchronous writing. If a writing task is in
	 * process, a wait instruction is issued before the lock is generated.
	 * 
	 * @throws InterruptedException An exception that indicates the waiting process has been
	 *             interrupted.
	 */
	private static synchronized void zipLock() throws InterruptedException
	{
		// As long as another process is writing
		while (writers > 0)
		{
			// Wait
			Util.class.wait();
		}
		// Update counter
		writers++;
	}

	/**
	 * Releases the lock on the zipping process that is used in order to prevent asynchronous
	 * writing. After the lock has been released, all waiting processes (if any) are notified.
	 */
	private static synchronized void zipUnlock()
	{
		// Update counter
		writers--;
		// Notify waiting processes
		Util.class.notifyAll();
	}

}